import re

from . import base
from . import primitives
from . import domains


# Patterns

REGEX_HEX = r'0-9a-f'
REGEX_PORT = r'[0-9]{2,5}'

REGEX_IPV4 = r'( ((%(octet)s\.){3} %(octet)s) )' % {
    'octet': '( 25[0-5] | 2[0-4][0-9] | [0-1]?[0-9]{1,2} )'
}

# 16 hex bytes
REGEX_IPV6_H16 = '[%s]{1,4}' % REGEX_HEX

# 16 hex bytes + traditional ip address
REGEX_IPV6_L32 = '(%(h16)s:%(h16)s|%(ipv4)s)' % {
    'h16': REGEX_IPV6_H16,
    'ipv4': REGEX_IPV4
}

# All allowed variations of H16 + L32
REGEX_IPV6 = r"""(
                                    (%(h16)s:){6}%(l32)s  |
                                ::  (%(h16)s:){5}%(l32)s  |
    (               %(h16)s )?  ::  (%(h16)s:){4}%(l32)s  |
    ( (%(h16)s:){,1}%(h16)s )?  ::  (%(h16)s:){3}%(l32)s  |
    ( (%(h16)s:){,2}%(h16)s )?  ::  (%(h16)s:){2}%(l32)s  |
    ( (%(h16)s:){,3}%(h16)s )?  ::  (%(h16)s:){1}%(l32)s  |
    ( (%(h16)s:){,4}%(h16)s )?  ::               %(l32)s  |
    ( (%(h16)s:){,5}%(h16)s )?  ::               %(h16)s  |
    ( (%(h16)s:){,6}%(h16)s )?  :: )""" % {
    'h16': REGEX_IPV6_H16,
    'l32': REGEX_IPV6_L32
}

REGEX_IP = r'^%s|%s$' % (REGEX_IPV4, REGEX_IPV6)

MAC_PATTERNS = {
    'hex': REGEX_HEX
}

REGEX_MAC = r"""(
    ^([%(hex)s]{2}[-]){5}([%(hex)s]{2})$  |
    ^([%(hex)s]{2}[:]){5}([%(hex)s]{2})$  |
    ^([%(hex)s]{12})                      |
    ^([%(hex)s]{6}[-:]([%(hex)s]{6}))$    |
    ^([%(hex)s]{4}(\.[%(hex)s]{4}){2})$
)""" % MAC_PATTERNS

_URL_GEN_DELIMS = set(':/?#[]@')
_URL_SUB_DELIMS = set('!$&\'()*+,;=')
_URL_ALL_DELIMS = _URL_GEN_DELIMS | _URL_SUB_DELIMS
_URL_UNRESERVED = set('-_.~')
_URL_PCHAR = _URL_SUB_DELIMS | _URL_UNRESERVED | set('%:@')
_URL_QUERY_EXTRAS = set('[]')  # nonstandard

_URL_VALID_CHARS = _URL_ALL_DELIMS | _URL_UNRESERVED | set('%')
_URL_VALID_CHAR_STRING = str.join('', _URL_VALID_CHARS)
_URL_UNSAFE_CHAR_STRING = '\x00-\x20<>{}|"`\\^\x7F-\x9F'


def u(allowed_chars):
    pattern = str.join('', _URL_VALID_CHARS - allowed_chars)

    pairs = [
        ('%', '%%'),
        (']', r'\]'),
        ('-', r'\-')
    ]
    for p in pairs:
        pattern = pattern.replace(p[0], p[1])

    return ('^' + _URL_UNSAFE_CHAR_STRING + pattern)


URI_PATTERNS = {
    'scheme': r'[%s]+' % ('A-Z0-9.+-'),
    'user': r'[%s]+' % u(_URL_UNRESERVED | _URL_SUB_DELIMS | set('%:')),
    'port': r'[0-9]{2,5}',
    'host4': REGEX_IPV4,
    'host6': r'[%s]+' % (REGEX_HEX + ':'),
    'hostn': r'[%s]+' % u(set('.-')),
    'path': r'[%s]*' % u(_URL_PCHAR | set('/')),
    'query': r'[%s]*' % u(_URL_PCHAR | set('/?') | _URL_QUERY_EXTRAS),
    'frag': r'[%s]*' % u(_URL_PCHAR | set('/?')),
}

REGEX_URL = r"""^(
        (?P<scheme> %(scheme)s ) ://
    (   (?P<user>   %(user)s   ) @   )?
    (\[ (?P<host6>  %(host6)s  ) ]
        | (?P<host4>  %(host4)s  )
        | (?P<hostn>  %(hostn)s  )     )
    ( : (?P<port>   %(port)s   )     )?
        (?P<path> / %(path)s   )?
    (\? (?P<query>  %(query)s  )     )?
    (\# (?P<frag>   %(frag)s   )     )?)$""" % URI_PATTERNS

EMAIL_PATTERNS = {
    'atext': '-A-Z0-9!#$%&\'*+/=?^_`{|}~',
    'qtext': '\x21\x23-\x5B\\\x5D-\x7E',
    'vchar': '\x21-\x7E'
}

REGEX_EMAIL = r"""^(
    ( ( [%(atext)s]+ (\.[%(atext)s]+)* )
      | ("( [%(qtext)s\s] | [\\%(vchar)s\s] )*") )
    @((?!-)[A-Z0-9-]{1,63}(?<!-)\.)+[A-Z]{2,63})$""" % EMAIL_PATTERNS


# Types

class IPAddressType(primitives.Primitive):
    NATIVE = str

    def __init__(self, regex=REGEX_IP, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.I + re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)


class IPv4Type(primitives.Primitive):
    NATIVE = str

    def __init__(self, regex=REGEX_IPV4, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.I + re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)


class IPv6Type(primitives.Primitive):
    NATIVE = str

    def __init__(self, regex=REGEX_IPV6, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.I + re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)


class MACAddressType(primitives.Primitive):
    NATIVE = str

    def __init__(self, regex=REGEX_MAC, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.I + re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)


class URLType(primitives.Primitive):
    NATIVE = str

    def __init__(self, regex=REGEX_URL, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.I + re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)


class EmailType(primitives.Primitive):
    NATIVE = str

    def __init__(self, regex=REGEX_EMAIL, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.I + re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)



from collections import OrderedDict

from . import base
from .. import cache
from .. import schematics
from .. import views


class RecordMeta(base.TypeMeta):
    def __new__(mcs, name, bases, namespace):
        # attribute accumulators
        fields = OrderedDict()
        field_functions = OrderedDict()
        validate_functions = OrderedDict()

        # gather values from base classes
        for b in reversed(bases):
            if hasattr(b, '_fields'):
                fields.update(b._fields)

            if hasattr(b, '_field_functions'):
                field_functions.update(b._field_functions)

            if hasattr(b, '_validate_functions'):
                validate_functions.update(b._validate_functions)

        # gather typerighter attributes
        for k, v in namespace.items():
            # collect type instances in fields dict
            if isinstance(v, base.Type):
                fields[k] = v

            # collect functions that generate fields
            elif k.startswith('field_') and callable(v):
                field_name = k[len('field_'):]
                field_functions[field_name] = v

            # collect validation functions
            elif k.startswith('validate_') and callable(v):
                validate_functions[k] = v

        # attach collected values
        namespace['_fields'] = fields
        namespace['_field_functions'] = field_functions
        namespace['_validate_functions'] = validate_functions

        # create the new type
        type_class = type.__new__(mcs, name, bases, namespace)

        # create schematic for type
        schematic = schematics.Schematic(type_class)
        setattr(type_class, '_schematic', schematic)

        # put type in cache
        cache.TypeCache().add(type_class)

        return type_class


class Record(base.Type, metaclass=RecordMeta):
    NATIVE = dict

    def __init__(
        self, strict=False, field_filters=None, export_nones=False, **kw
    ):
        super().__init__(**kw)
        self.strict = strict
        self.field_filters = field_filters
        self.export_nones = export_nones

    def __iter__(self):
        for field_name, type_instance in self._fields.items():
            yield field_name, type_instance

    def _filter(self, value, fields=None):
        for field_name, field_type in self:
            # Field name is in fields list and has a value
            if fields and field_name in fields and field_name in value:
                yield field_name, field_type
            # Field name has value
            elif field_name in value:
                yield field_name, field_type
            # Field can provide a value
            elif field_type.default:
                yield field_name, field_type

    def _convert(self, value, converter, fields=None):
        for fn, ti in self._filter(value, fields=fields):
            if fields and fn in fields and fn in value:
                v = converter(value[fn], ti)
                yield fn, v
            elif fn in value:
                v = converter(value[fn], ti)
                yield fn, v
            elif fn not in value and ti.default is not base.Unset:
                yield fn, ti.default

    @base.skip_falsy
    def to_primitive(self, value, **convert_args):
        converter = lambda field_value, ti: ti.to_primitive(field_value)
        return {
            k: v for k, v in self._convert(value, converter, **convert_args)
        }

    @base.skip_falsy
    def to_native(self, value, **convert_args):
        converter = lambda field_value, ti: ti.to_native(field_value)
        return {
            k: v for k, v in self._convert(value, converter, **convert_args)
        }

    def to_view(self, data=None):
        return views.make_view(self, data=data)

    @base.skip_falsy
    def validate_fields(self, value):
        for fn, ti in self:
            if fn in value:
                ti.validate(value[fn])
            else:
                ti.validate(base.Unset)



from typerighter import types


class Field(object):
    """A descriptor used to join a mutable View instance, that stores data,
    with the immutable Type instance, that only defines methods for operating
    on data.
    """
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, cls):
        if self.name in instance._data:
            return instance._data[self.name]
        return None

    def __set__(self, instance, value):
        instance._data[self.name] = value

    def __delete__(self, instance):
        del instance._data[self.name]


class View(object):
    """A View combines a `Record` with a dictionary to provide an object
    modeled after the record that can store data in a familiar object oriented
    manner.
    """
    def __init__(self, record, data=None):
        self._record = record
        self._data = data

    def to_primitive(self, **convert_args):
        return self._record.to_primitive(self._data, **convert_args)

    def to_native(self, **convert_args):
        return self._record.to_native(self._data, **convert_args)

    def validate(self):
        return self._record.validate(self._data)

    def __iter__(self):
        for k, v in self._data.items():
            if v is not types.Unset:
                yield k, v

    def __getattr__(self, name):
        if hasattr(self._record, name):
            return getattr(self._record, name)


def make_view(record, data=None):
    """Takes both a record and some data and produces View instance.

    :param Type record: The type that defines the view's shape
    :param dict data: Any initial data for the view's fields
    """
    if not data:
        data = {}

    # Wrap each field of the record in a `Field` instance
    attrs = {}
    for field_name, _ in record:
        attr = Field(field_name)
        attrs[field_name] = attr

    # Collect any factory setters
    for name in dir(record):
        if name.startswith('set_'):
            attr = getattr(record.__class__, name)
            attrs[name] = attr

    view_cls_name = '%sView' % (record.__class__.__name__)

    # Instantiate a View definition with the new fields
    RecordView = type(view_cls_name, (View,), attrs)

    return RecordView(record, data=data)



import re

from . import base
from . import domains
from .. import exceptions


class Primitive(base.Type):
    def __init__(self, choices=None, **kw):
        super().__init__(**kw)
        self.choices = choices

    def validate_choices(self, value):
        if self.choices and (value == base.Unset or value not in self.choices):
            e_msg = "Illegal choice for value: {}"
            raise exceptions.ValidationException(e_msg.format(value))


class BooleanType(Primitive):
    NATIVE = bool

    def to_primitive(self, value):
        if isinstance(value, bool):
            return value

        if isinstance(value, str):
            vl = value.lower()
            if vl == 'true':
                return True
            elif vl == 'false':
                return False

        if isinstance(value, (int, float)):
            return bool(value)

        e_msg = "Value is not a boolean: {}"
        raise exceptions.TypeException(e_msg.format(value))

    def validate_boolean(self, value):
        try:
            self.to_primitive(value)
        except exceptions.TypeException as te:
            ve = exceptions.ValidationException(*te.args)
            raise ve


class StringType(Primitive):
    NATIVE = str

    def __init__(self, max_length=None, min_length=None, regex=None, **kw):
        super().__init__(**kw)
        domains.LengthDomain(self, max_length, min_length)

        self.regex = regex

        if self.regex:
            try:
                compiled = re.compile(self.regex)
                self._regex = compiled
            except Exception:
                err_msg = "Regex failed to compile: {}"
                raise exceptions.TypeException(
                    err_msg.format(self.regex)
                )

    def is_falsy(self, value):
        if value is None or value == "":
            return True

        return super().is_falsy(value)

    def to_primitive(self, value):
        if isinstance(value, str):
            return value

        try:
            p = str(value)
            return p
        except Exception:
            err_msg = "Could not convert to string: {}"
            raise exceptions.TypeException(
                err_msg.format(err_msg.format(value))
            )

    @base.skip_falsy
    def validate_regex(self, value):
        if self.regex and not self._regex.match(value):
            err_msg = "Value did not match regex: {}"
            raise exceptions.ValidationException(
                err_msg.format(self, self.regex)
            )


class Number(Primitive):
    def __init__(self, max=None, min=None, **kw):
        super().__init__(**kw)
        domains.RangeDomain(self, max, min)


class IntegerType(Number):
    NATIVE = int


class FloatType(Number):
    NATIVE = float



import re

from typerighter import exceptions
from . import base


def skip_falsy(method):
    def wrapper(self, instance, value, *a, **kw):
        if self.is_falsy(instance, value):
            return value
        return method(self, instance, value, *a, **kw)
    return wrapper


class Domain:
    def __init__(self, instance):
        for attr_name in (a for a in dir(self) if a.startswith("validate_")):
            attr = getattr(self, attr_name)
            instance._validate_functions[attr_name] = attr

    def is_falsy(self, instance, value):
        return instance.is_falsy(value)


class RangeDomain(Domain):
    def __init__(self, instance, max=None, min=None):
        super().__init__(instance)
        instance.min = min
        instance.max = max

    def validate_min(self, instance, value):
        if not instance.min or value == base.Unset:
            return

        if instance.min and instance.is_falsy(value):
            err_msg = "Empty value not allowed with min: {}"
            raise exceptions.ValidationException(
                err_msg.format(instance.min)
            )

        if value < instance.min:
            err_msg = "Value below allowed min: {} < {}"
            raise exceptions.ValidationException(
                err_msg.format(value, instance.min)
            )

    def validate_max(self, instance, value):
        if not instance.max or instance.is_falsy(value):
            return

        if value > instance.max:
            err_msg = "Value exceeds allowed max: {} > {}"
            raise exceptions.ValidationException(
                err_msg.format(value, instance.max)
            )


class LengthDomain(Domain):
    def __init__(self, instance, max_length=None, min_length=None):
        super().__init__(instance)
        instance.max_length = max_length
        instance.min_length = min_length

    def validate_min_length(self, instance, value):
        if not instance.min_length or value == base.Unset:
            return

        if instance.min_length and instance.is_falsy(value):
            err_msg = "Empty value below min length: {}"
            raise exceptions.ValidationException(
                err_msg.format(instance.min_length)
            )

        if len(value) < instance.min_length:
            err_msg = "Value below min length: {} < {}"
            raise exceptions.ValidationException(
                err_msg.format(value, instance.min_length)
            )

    def validate_max_length(self, instance, value):
        if not instance.max_length or instance.is_falsy(value):
            return

        if len(value) > instance.max_length:
            err_msg = "Value length above max: {} > {}"
            raise exceptions.ValidationException(
                err_msg.format(value, instance.max_length)
            )


class RegexDomain(Domain):
    def __init__(self, instance, regex, *compile_args):
        super().__init__(instance)
        instance.regex = regex

        if instance.regex:
            try:
                compiled = re.compile(instance.regex, *compile_args)
                instance._regex = compiled
            except Exception:
                err_msg = "Regex failed to compile: {}"
                raise exceptions.TypeException(
                    err_msg.format(instance.regex)
                )

    def validate_regex(self, instance, value):
        if instance.is_falsy(value):
            return
        if not isinstance(value, str):
            return

        if instance._regex and not instance._regex.match(value):
            err_msg = "{} regex rejected value: {}"
            raise exceptions.ValidationException(
                err_msg.format(instance.__class__.__name__, value)
            )



import datetime
import re

from . import base
from . import primitives
from . import domains
from .. import exceptions


# Patterns

REGEX_FROM_ISO8601 = r"""(?P<year>\d{4})-(?P<month>\d\d)-(?P<day>\d\d)(?:T|\ )
    (?P<hour>\d\d):(?P<minute>\d\d)
    (?::(?P<second>\d\d)(?:(?:\.|,)(?P<sec_frac>\d{1,6}))?)?
    (?:(?P<tzd_offset>(?P<tzd_sign>[+−-])(?P<tzd_hour>\d\d):?(?P<tzd_minute>\d\d)?)
    |(?P<tzd_utc>Z))?$"""

REGEX_TO_ISO8601 = '%Y-%m-%dT%H:%M:%S.%f%z'

REGEX_FROM_TIME = r"""(?P<hour>\d\d):(?P<minute>\d\d)
    (?::(?P<second>\d\d)(?:(?:\.|,)(?P<sec_frac>\d{1,6}))?)?$"""


# Types

class DateTimeType(primitives.Primitive):
    NATIVE = datetime.datetime

    def __init__(self, regex=REGEX_FROM_ISO8601, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)

    @base.skip_falsy
    def to_native(self, value):
        if self.is_type_match(value):
            return value

        elif not isinstance(value, str):
            e_msg = "Value is not a string: {}"
            raise exceptions.TypeException(e_msg.format(value))

        # Verify regex
        match = self._regex.match(value)
        if not match:
            e_msg = "Value did not match date pattern: {}"
            raise exceptions.TypeException(e_msg.format(value))

        # Map of regex keys that had values
        parts = {k: v for k, v in match.groupdict().items() if v is not None}

        # Accessor for regex parts as numbers
        p = lambda name: int(parts.get(name, 0))

        # Microseconds
        microsecond = 0
        if 'sec_frac' in parts:
            sec_frac = p('sec_frac')
            sf_len = sec_frac and len(parts['sec_frac'])
            microsecond = sec_frac and sec_frac * 10 ** (6 - sf_len)

        # Timezones
        tz = None
        if 'tzd_utc' in parts:
            tz = datetime.timezone.utc
        elif 'tzd_offset' in parts:
            tz_sign = 1 if parts['tzd_sign'] == '+' else -1
            tz_offset = (p('tzd_hour') * 60 + p('tzd_minute')) * tz_sign
            if tz_offset == 0:
                tz = datetime.timezone.utc
            else:
                tz = datetime.timezone(datetime.timedelta(minutes=tz_offset))

        return self.NATIVE(
            p('year'), p('month'), p('day'), p('hour'), p('minute'),
            p('second'), microsecond, tz
        )

    @base.skip_falsy
    def to_primitive(self, value, context=None):
        if isinstance(value, str) and self._regex.match(value):
            return value
        elif isinstance(value, self.NATIVE):
            return value.strftime(REGEX_TO_ISO8601)
        e_msg = "Value must be iso8601 string or datetime.datetime: {}"
        raise exceptions.TypeException(e_msg.format(value))


class TimeType(primitives.Primitive):
    NATIVE = datetime.time

    def __init__(self, regex=REGEX_FROM_TIME, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)

    @base.skip_falsy
    def to_native(self, value):
        if isinstance(value, self.NATIVE):
            return value

        parts = {}
        match = self._regex.match(value)
        for k, v in match.groupdict().items():
            if v is not None:
                parts[k] = int(v)
        return datetime.time(**parts)

    @base.skip_falsy
    def to_primitive(self, value):
        if isinstance(value, self.NATIVE):
            return value.isoformat()
        if isinstance(value, str):
            return value
        e_msg = "Value must be time string or datetime.time: {}"
        raise exceptions.TypeException(e_msg.format(value))



"""
"""


from collections import OrderedDict
import inspect


def extract_argspec(klass):
    """Inspects a klass and creates a dict of keyword arguments and their
    default values.

    :param class klass: The class definition to inspect
    :return: a dictionary of default values found in argspec for class's init
    """
    argspec = OrderedDict()

    ka = inspect.getfullargspec(klass.__init__)
    if ka.defaults:
        for keyword, default_value in zip(ka.args[1:], ka.defaults):
            argspec[keyword] = default_value

    return argspec


def init_arg_capture(method):
    """A decorator that wraps a Type's `__init__` method for the purpose of
    capturing the arguments used when a Type is instantiated so it can then
    update the instance's argspec with what was actually used.
    """
    def wrapper(self, *a, **kw):
        method(self, *a, **kw)

        init_args = OrderedDict()
        init_args.update(self._schematic._argspec)
        for k, v in kw.items():
            init_args[k] = v

        self._schematic._init_args = init_args

    return wrapper


class Schematic(object):
    """A Schematic is a object that maintains a Type's argspec. It exists as a
    class to provide a namespace for relevant values.
    """
    def __init__(self, klass):
        """Initialize a Schematic instance with a Typerighter Type instance.

        :param class klass:
        """
        argspec = OrderedDict()

        # aggregate argspec from base classes
        base_classes = inspect.getmro(klass)
        for base in base_classes:
            if hasattr(base, '_schematic'):
                if hasattr(base._schematic, '_argspec'):
                    argspec.update(base._schematic._argspec)

        # add argspec
        k_argspec = extract_argspec(klass)
        argspec.update(k_argspec)
        self._argspec = argspec

        # this decorator must be applied *after* argspec is read
        klass.__init__ = init_arg_capture(klass.__init__)



class BaseException(Exception):
    pass


class UnsetException(BaseException):
    pass


class TypeException(BaseException):
    pass


class RecordException(BaseException):
    pass


class ValidationException(BaseException):
    pass


class ConfigException(BaseException):
    pass


class CacheMissException(BaseException):
    pass


class UncachableException(BaseException):
    pass



from collections import OrderedDict

from .. import cache
from .. import exceptions
from .. import schematics


class UnsetValue(object):
    """This class exists to put a label on the type of value that represents
    when a field does not _yet_ have a value.
    """
    pass


Unset = UnsetValue()


class TypeMeta(type):
    def __new__(mcs, name, bases, namespace):
        # attribute accumulators
        validate_functions = OrderedDict()

        # gather values from base classes
        for b in reversed(bases):
            if hasattr(b, '_validate_functions'):
                validate_functions.update(b._validate_functions)

        # gather typerighter attributes
        for k, v in namespace.items():
            if k.startswith('validate_') and callable(v):
                validate_functions[k] = v

        # attach collected values
        namespace['_validate_functions'] = validate_functions

        # create the new type
        type_class = type.__new__(mcs, name, bases, namespace)

        # create schematic for type
        schematic = schematics.Schematic(type_class)
        setattr(type_class, '_schematic', schematic)

        # put type in cache
        cache.TypeCache().add(type_class)

        return type_class


def skip_falsy(method):
    """A decorator that intercepts method calls to prevent falsy inputs from
    being validated unnecessarily.
    """
    def wrapper(self, value, *a, **kw):
        if self.is_falsy(value):
            return value
        return method(self, value, *a, **kw)
    return wrapper


class Type(object, metaclass=TypeMeta):
    """This class represents the top, and thus most ambiguous, point of the
    Typerighter hierarchy.

    It's purpose is to define the baseline expectations for every other `Type`
    in this library.
    """
    NATIVE = object  # identity function

    def __init__(self, default=Unset, required=False, strict=False):
        """Base initializer for Types. Implements the basic features for Types
        and provides the Type config options, such as setting a default value.

        :param object default: Any value you want to be used as a default
        inside `to_primitive` and `to_native`.
        :param bool required: If `True`, `Unset` values will trigger an
        exception during validation
        :param bool strict: Enforces all values are exactly the right type by
        validating without type coercion.
        """
        self.required = required
        self.strict = strict

        self.default = default
        if default != Unset:
            self.to_primitive = self._apply_default(self.to_primitive)
            self.to_native = self._apply_default(self.to_native)

    def is_falsy(self, value):
        """Checks a value and responds saying whether the `Type` considers it
        falsy.

        :param object value: The value to inspect
        :return: True or False
        """
        if value == Unset or value is None:
            return True
        return False

    def is_coercible(self, value):
        """Checks a value for whether or not it can be converted to the correct
        type. Falls back to the stricter `is_type_match` if `self.strict` is
        True.

        :param object value: The value to inspect
        """
        if self.strict:
            return self.is_type_match(value)
        try:
            self.to_native(value)
            return True
        except Exception:
            return False

    def is_type_match(self, value):
        """Checks if a value is an instance of this Type's native type.
        :param object value: The value to inspect
        """
        return isinstance(value, self.NATIVE)

    def to_schematic(self):
        """Returns a Type's Schematic
        """
        return (self.__class__.__name__, self._schematic.init_args)

    def _apply_default(self, method):
        def wrapper(value, *a, **kw):
            if value == Unset and self.default != Unset:
                return self.default
            return method(value, *a, **kw)
        return wrapper

    def to_primitive(self, value):
        """Converts a value to the primitive form of this type

        :param object value: The value to convert
        """
        return value

    def to_native(self, value):
        """Converts a value to the native form of this type

        :param object value: The value to convert
        """
        if self.is_falsy(value):
            return value

        if self.NATIVE != object:
            return self.NATIVE(value)

        return value

    def validate(self, value):
        """This validation function is the primary function responsible for
        calling all associated validators and for managing any details
        related to aggregation of validation results.

        :param object value: The value to convert
        """
        self._validate_required(value)
        self._validate_type_match(value)

        native = self.to_native(value)

        for func in self._validate_functions.values():
            func(self, native)

    def _validate_required(self, value):
        if self.required and value == Unset:
            e_msg = "Value required but not found"
            raise exceptions.ValidationException(e_msg)

    @skip_falsy
    def _validate_type_match(self, value):
        if self.strict and self.is_type_match(value):
            return
        elif self.is_coercible(value):
            return
        e_msg = "Value doesnt match type format {}".format(value)
        raise exceptions.ValidationException(e_msg)

    def __setattr__(self, name, value):
        """This method makes it impossible to overwrite any attributes
        that are subclasses of `Type`.
        """
        propobj = getattr(self.__class__, name, None)
        if propobj and isinstance(propobj, Type):
            return
        super().__setattr__(name, value)



import re

from . import base
from . import primitives
from . import domains


class UnixPathType(primitives.Primitive):
    NATIVE = str
    REGEX_UNIX_ABSPATH = r"^(\/[\w^ ]+)+\/?([\w.])+[^.]$"

    def __init__(self, regex=REGEX_UNIX_ABSPATH, **kw):
        super().__init__(**kw)
        domains.RegexDomain(self, regex, re.I + re.X)

    @base.skip_falsy
    def validate(self, value):
        super().validate(value)



from . import primitives
from . import domains
from .. import exceptions


class SumType(primitives.Primitive):
    def __init__(self, *types, **kwargs):
        super(SumType, self).__init__(**kwargs)
        self.types = types

    def _find_variant(self, value):
        for t in self.types:
            try:
                t.validate(value)
                return t
            except exceptions.ValidationException:
                pass
        else:
            err_msg = "No matching variant for value: {}"
            raise exceptions.TypeException(err_msg.format(value))

    def is_type_match(self, value):
        for t in self.types:
            if t.is_type_match(value):
                return True
        else:
            return False

    def to_schematic(self):
        variant_schematics = list()
        for t in self.types:
            schematic = t.to_schematic()
            variant_schematics.append(schematic)
        return (self.__class__.__name__, variant_schematics)

    def to_primitive(self, value):
        t = self._find_variant(value)
        return t.to_primitive(value)

    def to_native(self, value):
        t = self._find_variant(value)
        return t.to_native(value)

    def validate_by_variant_match(self, value):
        t = self._find_variant(value)
        if not t:
            e_msg = "Value did not pass any type checks: {}"
            raise exceptions.ValidationException(e_msg.format(value))


class Container(primitives.Primitive):
    def __init__(self, type, max_length=None, min_length=None, **kw):
        super().__init__(**kw)
        self.type = type
        domains.LengthDomain(self, max_length, min_length)

    def is_falsy(self, value):
        if value is None or super().is_falsy(value):
            return True

        # If calling len works, it's iterable
        try:
            if len(value) > 0:
                return False
        except Exception:
            pass

        return True

    def validate_items(self, value):
        err_msg = "Container is an abstract type %s"
        raise exceptions.ValidationException(err_msg % (self.type))


class ListType(Container):
    NATIVE = list

    def to_schematic(self):
        variant_schematics = list()
        for t in self.types:
            schematic = t.to_schematic()
            variant_schematics.append(schematic)
        return (self.__class__.__name__, variant_schematics)

    def to_primitive(self, value):
        if self.is_falsy(value):
            return value

        return [self.type.to_primitive(v) for v in value]

    def to_native(self, value):
        if self.is_falsy(value):
            return value

        return [self.type.to_native(v) for v in value]

    def validate_items(self, value):
        if self.is_falsy(value):
            return

        for v in value:
            try:
                self.type.validate(v)
                break
            except exceptions.ValidationException:
                pass
        else:
            e_msg = "No types in list match for item {}"
            raise exceptions.ValidationException(e_msg.format(value))



from .base import (
    Unset, TypeMeta, Type
)

from .primitives import (
    Primitive, BooleanType, IntegerType, FloatType, StringType
)

from .ids import (
    UUIDType
)

from .paths import (
    UnixPathType
)

from .timekeeping import (
    DateTimeType, TimeType
)

from .records import (
    RecordMeta, Record
)

from .composites import (
    SumType, Container, ListType
)

from .net import (
    IPAddressType, IPv4Type, IPv6Type, MACAddressType, URLType, EmailType
)

__all__ = [
    Unset, TypeMeta, Type,
    Primitive, BooleanType, IntegerType, FloatType, StringType,
    DateTimeType, TimeType,
    RecordMeta, Record,
    SumType, Container, ListType,
    IPAddressType, IPv4Type, IPv6Type, MACAddressType, URLType, EmailType
]



import inspect

from .exceptions import CacheMissException, UncachableException


class CacheMeta(type):
    """A metaclass for turning objects into singletons. Intended for use only
    with the `TypeCache`.
    """
    _instances = {}

    def __call__(cls, *a, **kw):
        if cls not in cls._instances:
            cls._instances[cls] = super(CacheMeta, cls).__call__(*a, **kw)
        return cls._instances[cls]


class TypeCache(object, metaclass=CacheMeta):
    """A store for every type defined inside some project. This store caches
    the classes that define types, but does not store any type instances.

    This cache is primarily useful to those who need scaffolding for types
    defined across multiple libraries.
    """
    def __init__(self):
        self._cache = {}

    def __str__(self):
        return repr(self)

    def items(self):
        return self._cache.items()

    def keys(self):
        return self._cache.keys()

    def get(self, name):
        if name in self._cache:
            return self._cache[name]
        else:
            err_msg = "Cache miss for type: %s" % (name)
            raise CacheMissException(err_msg)

    def add(self, item):
        type_name = item.__name__

        if not inspect.isclass(item):
            return False

        # By storing the first added `Type` as `_basetype`, we leverage the
        # convenience of `Type` being defined immediately after the first
        # metaclass to use `TypeCache`.
        #
        # As a result, we have access to `Type` without a circular import
        # error.
        if not hasattr(self, '_basetype'):
            self._basetype = item

        if self.cacheable(item):
            self._cache[type_name] = item
        else:
            err_msg = "Type not cachable: %s" % (type_name)
            raise UncachableException(err_msg)

    def cacheable(self, item):
        if not inspect.isclass(item):
            return False

        if issubclass(item, self._basetype):
            return True

        return False



from setuptools import setup, find_packages

from typerighter import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='typerighter',
    version=__version__,
    author='Jms Dnns',
    author_email='jdennis@gmail.com',
    description='Data Types for Cynical Humans',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/jmsdnns/typerighter',
    license='BSD',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-cov',
            'pycodestyle',
            'flake8',
            'tox'
        ], 
        'docs': [
            'sphinx',
            'sphinx_rtd_theme',
            'doc8'
        ]
    },
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only', 
    ]
)
